use crate::{
    error::{PERes, PIRes},
    index::{
        config::{IndexOrd, ValueMode},
        keeper::{IndexKeeper, IndexLimits, IndexModify},
        tree::nodes::{compare, Leaf, Node, PageIter, PageIterBack, TreeNode, TreeNodeRef, Value},
    },
};
use std::{cmp::Ordering, fmt::Display, ops::Bound, rc::Rc};

use self::path::{BottomDepth, Path, PathItem, Position, TopDepth};
use super::config::IndexTypeInternal;

pub mod nodes;
#[cfg(test)]
mod nodes_tests;
mod path;
#[cfg(test)]
mod tests;

#[derive(Clone, Eq, PartialEq, Debug)]
pub enum ValueChange<V> {
    Add(V),
    Remove(Option<V>),
}

pub struct KeyChanges<K, V> {
    pub k: K,
    changes: Vec<ValueChange<V>>,
}
impl<K: Clone, V: Clone> Clone for KeyChanges<K, V> {
    fn clone(&self) -> Self {
        Self {
            k: self.k.clone(),
            changes: self.changes.clone(),
        }
    }
}

impl<K: IndexOrd + Clone, V: IndexOrd + Clone> KeyChanges<K, V> {
    pub fn new(k: K, changes: Vec<ValueChange<V>>) -> KeyChanges<K, V> {
        KeyChanges { k, changes }
    }
}

#[cfg(not(feature = "index_container_static"))]
impl<K: IndexOrd + Clone, V: IndexOrd + Clone> KeyChanges<K, V> {
    pub fn put(&mut self, v: V) {
        self.changes.push(ValueChange::Add(v))
    }

    pub fn remove(&mut self, v: Option<V>) {
        self.changes.push(ValueChange::Remove(v))
    }

    pub fn changes(&self) -> Vec<ValueChange<V>> {
        self.changes.clone()
    }
}

#[cfg(test)]
impl<K: IndexOrd + Clone, V: IndexOrd + Clone> KeyChanges<K, V> {
    pub fn single_add(k: K, v: V) -> KeyChanges<K, V> {
        KeyChanges {
            k,
            changes: vec![ValueChange::Add(v)],
        }
    }

    fn single_delete(k: K, v: Option<V>) -> KeyChanges<K, V> {
        KeyChanges {
            k,
            changes: vec![ValueChange::Remove(v)],
        }
    }
}

struct ChangeResults {
    add: Option<usize>,
    remove: Option<usize>,
}

impl<K: IndexTypeInternal, V: IndexOrd + Clone> KeyChanges<K, V> {
    fn apply(&self, leaf: &mut Leaf<K, V>, value_mode: ValueMode, index: &str) -> PIRes<ChangeResults> {
        let mut cr = ChangeResults {
            add: None,
            remove: None,
        };
        for vc in &self.changes {
            match vc {
                ValueChange::Add(v) => {
                    leaf.add(&self.k, v, value_mode.clone(), index)?;
                    cr.add = if let Some(add) = cr.add { Some(add + 1) } else { Some(1) };
                }
                ValueChange::Remove(r) => {
                    if leaf.remove(&self.k, r) {
                        cr.remove = if let Some(remove) = cr.remove {
                            Some(remove + 1)
                        } else {
                            Some(1)
                        };
                    }
                }
            };
        }
        Ok(cr)
    }
}

pub trait Index<K, V> {
    fn get(&mut self, k: &K) -> PERes<Option<Value<V>>>;
    fn iter_from(&mut self, first: Bound<&K>) -> PERes<PageIter<K, V>>;
    fn back_iter_from(&mut self, first: Bound<&K>) -> PERes<PageIterBack<K, V>>;
}

pub trait IndexApply<K, V>: Index<K, V> {
    fn apply(&mut self, adds: &[KeyChanges<K, V>]) -> PIRes<()>;
}
impl<K: Display> Display for PosRef<K> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}:{}=>{}", self.k, self.pos, self.node_ref)
    }
}

#[derive(PartialEq, Clone, Debug, Eq)]
pub struct PosRef<K> {
    pub k: K,
    pub pos: usize,
    pub node_ref: TreeNodeRef,
}

impl<K: Clone> PosRef<K> {
    pub(crate) fn new(k: &K, pos: usize, node_ref: TreeNodeRef) -> PosRef<K> {
        PosRef {
            k: k.clone(),
            pos,
            node_ref,
        }
    }
}

fn split_and_save<K, V, I>(
    index: &mut I,
    id: TreeNodeRef,
    node: Rc<TreeNode<K, V>>,
    version: u16,
) -> PIRes<Vec<(K, TreeNodeRef)>>
where
    K: IndexOrd + Clone,
    I: IndexModify<K, V>,
{
    let (mut first, _) = index.owned(&id, node);
    let new_nodes = first.split(index.limits().top());
    let mut ids = Vec::new();
    for (k, new_node) in new_nodes {
        let saved_id = index.insert(new_node)?;
        ids.push((k, saved_id));
    }
    index.update(&id, first, version)?;
    Ok(ids)
}

fn recompute_path<K, V, I>(index: &mut I, child: &ChildChanged<K>) -> PIRes<Path<K>>
where
    K: IndexOrd + Clone,
    I: IndexModify<K, V>,
{
    let node = index.get_root_refresh()?.expect("Root is there");
    let mut path = Path::new(child.search_key().clone());
    let mut cur_node = PosRef::new(child.search_key(), 0, node);
    loop {
        let mut restart = true;
        if let Some((node_modify, version)) = index.load_modify(&cur_node.node_ref)? {
            path.push(cur_node.clone(), version, node_modify.len());
            if cur_node.node_ref == child.item.pos_ref().node_ref {
                path.compute_depths_start(child.item.bottom_depth());
                path.replace_search_key(cur_node.k);
                break Ok(path);
            }
            if let TreeNode::Node(n) = &*node_modify {
                if let Some(x) = n.find_write(&child.search_key()) {
                    cur_node = x;
                    restart = false;
                }
            }
        }
        if restart {
            let node = index.get_root_refresh()?.expect("Root is there");
            cur_node = PosRef::new(&child.search_key(), 0, node);
            path.clear();
        }
    }
}

fn recompute_path_depth<K, V, I>(index: &mut I, depth: BottomDepth, k: &K) -> PIRes<Path<K>>
where
    K: IndexOrd + Clone,
    I: IndexModify<K, V>,
{
    let node = index.get_root_refresh()?.expect("Root is there");
    let mut path = Path::new(k.clone());
    let mut cur_node = PosRef::new(k, 0, node);
    loop {
        let mut restart = true;
        if let Some((node_modify, version)) = index.load_modify(&cur_node.node_ref)? {
            path.push(cur_node.clone(), version, node_modify.len());
            match &*node_modify {
                TreeNode::Node(n) => {
                    if let Some(x) = n.find_write(k) {
                        cur_node = x;
                        restart = false;
                    }
                }
                TreeNode::Leaf(_n) => {
                    path.compute_depths();
                    break;
                }
            }
        }
        if restart {
            let node = index.get_root_refresh()?.expect("Root is there");
            cur_node = PosRef::new(k, 0, node);
            path.clear();
        }
    }
    path.short_to_depth(depth);
    Ok(path)
}

fn recompute_path_depth_pre<K, V, I>(index: &mut I, depth: BottomDepth, k: &K) -> PIRes<Path<K>>
where
    K: IndexOrd + Clone,
    I: IndexModify<K, V>,
{
    let node = index.get_root_refresh()?.expect("Root is there");
    let mut path = Path::new(k.clone());
    let mut cur_node = PosRef::new(k, 0, node);
    loop {
        let mut restart = true;
        if let Some((node_modify, version)) = index.load_modify(&cur_node.node_ref)? {
            path.push(cur_node.clone(), version, node_modify.len());
            match &*node_modify {
                TreeNode::Node(n) => {
                    if let Some(x) = n.find_pre_write(k) {
                        cur_node = x;
                        restart = false;
                    }
                }
                TreeNode::Leaf(_n) => {
                    path.replace_search_key(cur_node.k.clone());
                    path.compute_depths();
                    break;
                }
            }
        }
        if restart {
            let node = index.get_root_refresh()?.expect("Root is there");
            cur_node = PosRef::new(k, 0, node);
            path.clear();
        }
    }
    path.short_to_depth(depth);
    Ok(path)
}

fn lock_parents<K, V, I>(
    index: &mut I,
    updates: Vec<LockGroup<K>>,
) -> PIRes<(Vec<LockGroup<K>>, Vec<Vec<ToMergeCheck<K>>>)>
where
    K: IndexTypeInternal,
    V: IndexTypeInternal,
    I: IndexModify<K, V>,
{
    let mut group_changes = LockGroups::new();
    let mut checks_groups = Vec::new();
    for update in &updates {
        let mut parent_group = update.parent_group(index.limits());
        let mut group_len = parent_group.len();
        let mut cur = 0;
        let mut checks = Vec::new();
        let mut retry = 0;
        while cur < group_len {
            let recompute = {
                let parent = &mut parent_group[cur];

                if !parent.to_lock() {
                    cur += 1;
                    continue;
                }
                let last_pos = parent.path.last().unwrap();
                let pos_ref = last_pos.pos_ref();
                let node_ref = pos_ref.node_ref;
                if let Some((node, version)) = index.load_modify(&node_ref)? {
                    if last_pos.version() == version {
                        let path = parent.path.clone();
                        if lock_logic(index, path, node, &mut group_changes, Some(parent))? {
                            checks.extend(parent.children_checks());
                            cur += 1;
                            retry = 0;
                            false
                        } else {
                            true
                        }
                    } else {
                        true
                    }
                } else {
                    true
                }
            };

            if recompute {
                retry += 1;
                if retry > 1000000 {
                    return Err(crate::IndexChangeError::ReachedLimitOfRetry);
                }
                let mut children = Vec::<ChildChanged<K>>::new();
                let mut inserted = 0;
                let mut prev_path: Option<Path<K>> = None;
                let old_children = parent_group[cur].children.clone();
                for mut child in old_children {
                    let mut path = recompute_path(index, &child)?;
                    child.item = path.pop().unwrap();
                    if let Some(pp) = prev_path {
                        if pp.last_id() != path.last_id() {
                            if inserted == 0 {
                                parent_group[cur].replace(pp.clone(), children);
                            } else {
                                let element = ParentChange::new_childs(pp.clone(), children);
                                group_len += 1;
                                parent_group.insert(cur + inserted, element);
                            }
                            inserted += 1;
                            children = Vec::new();
                        }
                    }
                    prev_path = Some(path);
                    children.push(child);
                }
                if !children.is_empty() {
                    if let Some(pp) = prev_path {
                        if inserted == 0 {
                            parent_group[cur].replace(pp, children);
                        } else {
                            let element = ParentChange::new_childs(pp, children);
                            group_len += 1;
                            parent_group.insert(cur + inserted, element);
                        }
                    }
                }
            }
        }
        if !checks.is_empty() {
            checks_groups.push(checks);
        }
    }
    group_changes.finish();
    Ok((group_changes.groups, checks_groups))
}

#[derive(Clone)]
struct ToMergeCheck<K> {
    parent: PosRef<K>,
    child: PosRef<K>,
    check_key: bool,
}

impl<K> ToMergeCheck<K> {
    fn new(parent: PosRef<K>, child: PosRef<K>, check_key: bool) -> Self {
        Self {
            parent,
            child,
            check_key,
        }
    }
}

fn rebalance<K, V, I>(index: &mut I, to_balance: Vec<ToMergeCheck<K>>) -> PIRes<()>
where
    K: IndexOrd + Clone + Display,
    I: IndexModify<K, V>,
    V: Display,
{
    let mut prev: Option<(ToMergeCheck<K>, Rc<TreeNode<K, V>>, u16)> = None;
    for rmc in &to_balance {
        let pos = &rmc.child;
        if let Some((node, version)) = index.load_modify(&pos.node_ref)? {
            if let Some((prev_id, prev_node, prev_version)) = prev {
                if node.len() < index.limits().bottom() || prev_node.len() < index.limits().bottom() {
                    let (mut dest_merge, _) = index.owned(&prev_id.child.node_ref, prev_node);
                    let (mut source_merge, _) = index.owned(&pos.node_ref, node);
                    let merge_key = source_merge.get_prev().as_ref().unwrap().clone();
                    //let parent_key = &pos.k;
                    dest_merge.merge_right(&merge_key, &mut source_merge);
                    let parent_id = rmc.parent.node_ref;

                    let (parent_node, parent_version) = index.load_modify(&parent_id)?.unwrap();
                    let mut parent = index.owned(&parent_id, parent_node).0.as_node();
                    let parent_key = parent.find_key(&pos.node_ref).unwrap().clone();
                    let next_to_set = parent.remove_key(&parent_key);
                    index.update(&parent_id, TreeNode::Node(parent), parent_version)?;

                    if let Some(next) = next_to_set {
                        let prev_parent_id = prev_id.parent.node_ref;
                        let (prev_parent_node, prev_parent_version) = index.load_modify(&prev_parent_id)?.unwrap();
                        let mut prev_parent = index.owned(&prev_parent_id, prev_parent_node).0.as_node();
                        prev_parent.swap_next(&merge_key, &next);
                        index.update(&prev_parent_id, TreeNode::Node(prev_parent), prev_parent_version)?;
                    }
                    index.delete(&pos.node_ref, version)?;
                    index.update(&prev_id.child.node_ref, dest_merge, prev_version)?;
                    let (node, version) = index.load_modify(&prev_id.child.node_ref)?.unwrap();
                    prev = Some((prev_id, node, version));
                } else if rmc.check_key {
                    let parent_id = rmc.parent.node_ref;
                    let (parent_node, parent_version) = index.load_modify(&parent_id)?.unwrap();

                    let old_key = if let TreeNode::Node(parent_node_n) = &*parent_node {
                        parent_node_n.find_key(&pos.node_ref)
                    } else {
                        unreachable!()
                    }
                    .unwrap()
                    .clone();
                    let new_key = node.get_prev().as_ref().unwrap();
                    if compare(&old_key, new_key) != Ordering::Equal {
                        let (need_change, need_change_prev_next) = if let TreeNode::Node(path_parent_n) = &*parent_node
                        {
                            path_parent_n.need_swap_key(&old_key)
                        } else {
                            (false, false)
                        };
                        if need_change || need_change_prev_next {
                            let mut parent = index.owned(&parent_id, parent_node).0.as_node();
                            parent.swap_key(&old_key, new_key);
                            index.update(&parent_id, TreeNode::Node(parent), parent_version)?;
                        }

                        let prev_parent_id = prev_id.parent.node_ref;
                        let (prev_parent_node, prev_parent_version) = index.load_modify(&prev_parent_id)?.unwrap();

                        let need_change_next = if let TreeNode::Node(path_parent_n) = &*prev_parent_node {
                            path_parent_n.need_swap_next(&old_key)
                        } else {
                            false
                        };
                        debug_assert_eq!(need_change_prev_next, need_change_next);
                        if need_change_next {
                            let mut prev_parent = index.owned(&prev_parent_id, prev_parent_node).0.as_node();
                            prev_parent.swap_next(&old_key, new_key);
                            index.update(&prev_parent_id, TreeNode::Node(prev_parent), prev_parent_version)?;
                        }
                    }
                    prev = Some((rmc.clone(), node, version));
                } else {
                    prev = Some((rmc.clone(), node, version));
                }
            } else {
                prev = Some((rmc.clone(), node, version));
            }
        }
    }
    if to_balance.len() == 1 {
        let tb = &to_balance[0];
        if tb.check_key {
            let parent_id = tb.parent.node_ref;
            let (parent_node, parent_version) = index.load_modify(&parent_id)?.unwrap();

            let old_key = if let TreeNode::Node(parent_node_n) = &*parent_node {
                parent_node_n.find_key(&tb.child.node_ref)
            } else {
                unreachable!()
            }
            .unwrap()
            .clone();
            let (node, _version) = index.load_modify(&tb.child.node_ref)?.unwrap();
            let new_key = node.get_prev().as_ref().unwrap();
            if compare(&old_key, new_key) != Ordering::Equal {
                let (need_change, need_change_prev_next) = if let TreeNode::Node(path_parent_n) = &*parent_node {
                    path_parent_n.need_swap_key(&old_key)
                } else {
                    (false, false)
                };
                if need_change || need_change_prev_next {
                    let mut parent = index.owned(&parent_id, parent_node).0.as_node();
                    parent.swap_key(&old_key, new_key);
                    index.update(&parent_id, TreeNode::Node(parent), parent_version)?;
                }
                debug_assert!(!need_change_prev_next);
            }
        }
    }

    for rmc in &to_balance {
        let pos = &rmc.child;

        if let Some((node, version)) = index.load_modify(&pos.node_ref)? {
            if node.len() > index.limits().top() {
                let (mut first, _) = index.owned(&pos.node_ref, node);
                let new_nodes = first.split(index.limits().top());
                let mut ids = Vec::new();
                for (k, new_node) in new_nodes {
                    let saved_id = index.insert(new_node)?;
                    ids.push((k, saved_id));
                }
                index.update(&pos.node_ref, first, version)?;

                let rp = rmc.parent.node_ref;
                let (parent_node, parent_version) = index.load_modify(&rp)?.unwrap();
                let mut parent = index.owned(&rp, parent_node).0.as_node();
                parent.insert_after_key(&ids);
                index.update(&rp, TreeNode::Node(parent), parent_version)?;
            }
        }
    }
    Ok(())
}

#[derive(Clone)]
enum Op {
    Add(usize),
    Remove,
    Nothing,
}

#[derive(Clone)]
struct ChildChanged<K> {
    item: PathItem<K>,
    search_key: K,
    operation: Op,
    check_key: bool,
}
impl<K: Clone> ChildChanged<K> {
    fn new(item: PathItem<K>, search_key: K, operation: Op, check_key: bool) -> Self {
        Self {
            item,
            search_key,
            operation,
            check_key,
        }
    }

    pub fn to_check(&self, parent: PosRef<K>) -> ToMergeCheck<K> {
        ToMergeCheck::new(
            parent,
            self.item.pos_ref().clone(),
            self.check_key || self.item.position() == Position::First,
        )
    }
    pub fn search_key(&self) -> &K {
        &self.search_key
    }
}
struct ParentChange<K> {
    path: Path<K>,
    children: Vec<ChildChanged<K>>,
    remove: usize,
    add: usize,
    change_prev_key: bool,
    change_next_key: bool,
}
impl<K: IndexOrd + Clone> ParentChange<K> {
    fn new_childs(path: Path<K>, children: Vec<ChildChanged<K>>) -> Self {
        debug_assert!(!path.is_empty());
        let mut add = 0;
        let mut remove = 0;
        let mut is_same_prev_key = false;
        let mut is_same_next_key = false;
        for child in &children {
            if path.last_position() == Position::First || child.check_key {
                is_same_prev_key = true;
            }
            if child.item.position() == Position::Last {
                is_same_next_key = true;
            }
            match child.operation {
                Op::Add(counts) => add += counts,
                Op::Remove => remove += 1,
                Op::Nothing => (),
            }
        }
        Self {
            path,
            children,
            remove,
            add,
            change_prev_key: is_same_prev_key,
            change_next_key: is_same_next_key,
        }
    }
    fn replace(&mut self, path: Path<K>, children: Vec<ChildChanged<K>>) {
        debug_assert!(!path.is_empty());
        let mut add = 0;
        let mut remove = 0;
        let mut is_same_prev_key = false;
        let mut is_same_next_key = false;
        for child in &children {
            if child.item.position() == Position::First || child.check_key {
                is_same_prev_key = true;
            }
            if child.item.position() == Position::Last {
                is_same_next_key = true;
            }
            match child.operation {
                Op::Add(counts) => add += counts,
                Op::Remove => remove += 1,
                Op::Nothing => (),
            }
        }
        self.path = path;
        self.children = children;
        self.add = add;
        self.remove = remove;
        self.change_prev_key = is_same_prev_key;
        self.change_next_key = is_same_next_key;
    }
    fn new_child(child: &LockItem<K>, limits: &IndexLimits) -> Self {
        let mut path = child.path.clone();
        path.pop();
        debug_assert!(!path.is_empty());
        let mut inst = Self {
            path,
            children: Vec::new(),
            add: 0,
            remove: 0,
            change_prev_key: false,
            change_next_key: false,
        };
        inst.add_child(child, limits);
        inst
    }
    pub fn add_child(&mut self, child: &LockItem<K>, limits: &IndexLimits) {
        if child.path.last_position() == Position::First || child.check_key {
            self.change_prev_key = true;
            true
        } else {
            false
        };
        let check_key = child.check_key;
        if child.path.last_position() == Position::Last {
            self.change_next_key = true;
        }
        let operation = if child.len() < limits.bottom() {
            self.remove += 1;
            Op::Remove
        } else if child.len() > limits.top() {
            let add = nodes::block_counts_for_split(child.len(), limits.top());
            self.add += add;
            Op::Add(add)
        } else {
            Op::Nothing
        };
        self.children.push(ChildChanged::new(
            child.path.last().cloned().unwrap(),
            child.path.search_key().clone(),
            operation,
            check_key,
        ));
    }

    pub fn to_lock(&self) -> bool {
        self.add != 0 || self.remove != 0 || self.change_prev_key || self.change_next_key
    }
    pub fn id(&self) -> TreeNodeRef {
        self.path.last_id().unwrap()
    }

    pub fn children_checks(&self) -> Vec<ToMergeCheck<K>> {
        let parent = self.path.last_pos_ref().unwrap();
        self.children
            .iter()
            .map(|i| i.to_check(parent.clone()))
            .collect::<Vec<_>>()
    }
}

#[derive(Clone)]
struct LockItem<K: Clone> {
    path: Path<K>,
    len: usize,
    add: Option<usize>,
    removes: Option<usize>,
    check_key: bool,
}

impl<K: IndexOrd + Clone> LockItem<K> {
    fn new(path: Path<K>, len: usize) -> Self {
        Self {
            path,
            len,
            add: None,
            removes: None,
            check_key: false,
        }
    }

    fn new_changes(path: Path<K>, len: usize, parent_change: &ParentChange<K>) -> Self {
        Self {
            path,
            len,
            add: Some(parent_change.add),
            removes: Some(parent_change.remove),
            check_key: parent_change.change_prev_key,
        }
    }

    pub fn merge_v(&mut self, parent_change: &ParentChange<K>, path: Path<K>) {
        self.add = Some(self.add.unwrap_or(0) + parent_change.add);

        self.removes = Some(self.removes.unwrap_or(0) + parent_change.remove);

        self.check_key |= parent_change.change_prev_key;
        self.path = path;
    }
    pub fn len(&self) -> usize {
        let mut len = self.len;
        if let Some(a) = self.add {
            len += a;
        }
        if let Some(r) = self.removes {
            len -= r;
        }
        len
    }

    pub fn parent_id(&self) -> Option<TreeNodeRef> {
        self.path
            .get(TopDepth(self.path.len() - 2))
            .map(|x| x.pos_ref().node_ref)
    }
}

struct LockGroup<K: Clone> {
    depth: BottomDepth,
    group: Vec<LockItem<K>>,
}

impl<K: IndexOrd + Clone> LockGroup<K> {
    fn new(depth: BottomDepth) -> Self {
        Self {
            depth,
            group: Vec::new(),
        }
    }
    pub fn is_for(&self, node_ref: &TreeNodeRef) -> bool {
        self.group.is_empty()
            || self
                .group
                .iter()
                .rfind(|e| e.path.last_id() == Some(*node_ref))
                .is_some()
    }
    pub fn add(&mut self, mut path: Path<K>, len: usize, pc: &Option<&ParentChange<K>>) {
        path.compute_depths_start(self.depth);
        for g in &mut self.group {
            if g.path.point_same_node(&path) {
                if let Some(parent_change) = pc {
                    if parent_change.path.point_same_node(&path) {
                        g.merge_v(parent_change, path);
                    }
                }
                return;
            }
        }

        if let Some(parent_change) = pc {
            if parent_change.path.point_same_node(&path) {
                self.group.push(LockItem::new_changes(path, len, parent_change));
            } else {
                self.group.push(LockItem::new(path, len));
            }
        } else {
            self.group.push(LockItem::new(path, len));
        }
    }

    pub fn set_change(&mut self, nr: &TreeNodeRef, cr: ChangeResults) {
        let mut found = false;
        for g in &mut self.group {
            if g.path.last_pos_ref().map(|r| &r.node_ref) == Some(nr) {
                if let Some(nadd) = cr.add {
                    g.add = Some(if let Some(add) = g.add { add + nadd } else { nadd });
                }
                if let Some(nrem) = cr.remove {
                    g.removes = Some(if let Some(r) = g.removes { r + nrem } else { nrem });
                }
                found = true;
                break;
            }
        }
        assert!(found);
    }

    pub fn parent_group(&self, limits: &IndexLimits) -> Vec<ParentChange<K>> {
        let mut parents = Vec::new();

        let mut parent_change: Option<ParentChange<K>> = None;
        for value in &self.group {
            parent_change = Some(if let Some(mut pc) = parent_change {
                if pc.id() == value.parent_id().unwrap() {
                    pc.add_child(value, limits);
                    pc
                } else {
                    if !pc.children.is_empty() {
                        parents.push(pc);
                    }
                    ParentChange::new_child(value, limits)
                }
            } else {
                ParentChange::new_child(value, limits)
            });
        }
        if let Some(x) = parent_change {
            if !x.children.is_empty() {
                parents.push(x);
            }
        }

        parents
    }
}

struct ToLock<K> {
    prev: Option<K>,
    next: Option<K>,
    len: usize,
    path: Path<K>,
}

struct ToLockNodes<K> {
    to_lock: Vec<ToLock<K>>,
}
impl<K: IndexOrd + Clone> ToLockNodes<K> {
    fn new() -> Self {
        Self {
            to_lock: Default::default(),
        }
    }
    fn push<V>(&mut self, node: &Rc<TreeNode<K, V>>, path: Path<K>) {
        debug_assert!(path.last_bottom_depth().is_some());
        self.to_lock.push(ToLock {
            prev: node.get_prev().clone(),
            next: node.get_next().clone(),
            len: node.len(),
            path,
        });
    }
    fn are_one_after_the_other(&self) -> bool {
        let mut prev_next: Option<Option<K>> = None;
        for to_lock in &self.to_lock {
            if let Some(pr) = prev_next {
                if let (Some(f), Some(s)) = (pr, &to_lock.prev) {
                    if f.cmp(s) != Ordering::Equal {
                        return false;
                    }
                } else {
                    return false;
                }
            }
            prev_next = Some(to_lock.next.clone());
        }
        true
    }
    fn lock_all<I: IndexModify<K, V>, V>(&self, index: &mut I) -> PIRes<bool> {
        let mut locked = Vec::new();
        for to_lock in &self.to_lock {
            let last = to_lock.path.last().unwrap();
            if !index.lock(&last.pos_ref().node_ref, last.version())? {
                for to_unlock in locked {
                    index.unlock(&to_unlock);
                }
                return Ok(false);
            }
            locked.push(last.pos_ref().node_ref);
        }
        Ok(true)
    }
    fn len(&self) -> usize {
        self.to_lock.len()
    }
}

/// lock the node and a sibling node that may be target by a merge with the current node.
fn lock_logic<K, V, I>(
    index: &mut I,
    path: Path<K>,
    node: Rc<TreeNode<K, V>>,
    lock_groups: &mut LockGroups<K>,
    pc: Option<&ParentChange<K>>,
) -> PIRes<bool>
where
    K: IndexTypeInternal,
    V: IndexTypeInternal,
    I: IndexModify<K, V>,
{
    let mut bottom_limit = index.limits().bottom() as isize;
    let top_depth = path.depth();
    let cur_item = path.last().unwrap();
    let cur_node = cur_item.pos_ref();
    if !node.check_range(path.search_key()) {
        return Ok(false);
    }

    let bottom_depth = path.last_bottom_depth().unwrap();

    if top_depth > TopDepth(0) {
        let parent_item = path.get(top_depth - TopDepth(1)).unwrap();
        let parent = parent_item.pos_ref();
        let mut to_lock = ToLockNodes::new();
        if let Some((n, version)) = index.load_modify(&parent.node_ref)? {
            if parent_item.version() != version {
                return Ok(false);
            }
            if let TreeNode::Node(parent_node) = &*n {
                let mut len = node.len() as isize;
                if let Some(parent_change) = pc {
                    len += parent_change.add as isize;
                    len -= parent_change.remove as isize;
                    if parent_change.change_prev_key {
                        if let Some(key) = node.get_prev() {
                            let pre_path = recompute_path_depth_pre(index, bottom_depth, key)?;
                            let pr = pre_path.last_pos_ref().unwrap();
                            if let Some((n, _v)) = index.load_modify(&pr.node_ref)? {
                                let len = n.len() as isize;
                                to_lock.push(&n, pre_path);
                                bottom_limit -= len;
                            } else {
                                return Ok(false);
                            }
                        }
                    }
                }
                bottom_limit -= len;

                to_lock.push(&node, path.clone());
                let iter = if let Some(iter) = parent_node.iter_from(Some(cur_node)) {
                    iter
                } else {
                    return Ok(false);
                };
                for next_pos in iter {
                    if let Some((n, v)) = index.load_modify(&next_pos.node_ref)? {
                        let len = n.len() as isize;
                        let mut new_path = path.clone();
                        new_path.pop();
                        new_path.replace_search_key(next_pos.k.clone());
                        new_path.push(next_pos, v, n.len());
                        new_path.compute_depths_start(bottom_depth);
                        to_lock.push(&n, new_path);
                        bottom_limit -= len;
                    } else {
                        return Ok(false);
                    }
                    if bottom_limit <= 0 {
                        break;
                    }
                }

                let mut next_key = parent_node.get_next().clone();
                while (bottom_limit > 0 || to_lock.len() < 2) && next_key.is_some() {
                    if let Some(key) = next_key.take() {
                        let mut parent_path = recompute_path_depth(index, bottom_depth, &key)?;
                        parent_path.pop();
                        let parent_id = &parent_path.last_pos_ref().unwrap();
                        if let Some((p, _)) = index.load_modify(&parent_id.node_ref)? {
                            if let TreeNode::Node(parent) = &*p {
                                let iter = if let Some(iter) = parent.iter_from(None) {
                                    iter
                                } else {
                                    return Ok(false);
                                };
                                for next_pos in iter {
                                    if let Some((n, v)) = index.load_modify(&next_pos.node_ref)? {
                                        let len = n.len() as isize;
                                        let mut new_path = parent_path.clone();
                                        new_path.replace_search_key(next_pos.k.clone());
                                        new_path.push(next_pos, v, n.len());
                                        new_path.compute_depths_start(bottom_depth);
                                        to_lock.push(&n, new_path);
                                        bottom_limit -= len;
                                    } else {
                                        return Ok(false);
                                    }
                                    if bottom_limit <= 0 {
                                        break;
                                    }
                                }
                                next_key = p.get_next().clone();
                            }
                        }
                    }
                }
            } else {
                return Ok(false);
            }
        } else {
            return Ok(false);
        };

        if !to_lock.are_one_after_the_other() {
            return Ok(false);
        }
        if !to_lock.lock_all(index)? {
            return Ok(false);
        }
        let first_path = &to_lock.to_lock.first().unwrap().path;
        let lock_group = lock_groups.group_for_id(first_path.last_pos_ref().unwrap(), bottom_depth);
        for to_lock in to_lock.to_lock {
            lock_group.add(to_lock.path, to_lock.len, &pc);
        }
        Ok(true)
    } else if index.lock(&cur_node.node_ref, cur_item.version())? {
        let lock_group = lock_groups.group_for_id(cur_node, bottom_depth);
        lock_group.add(path, node.len(), &pc);
        Ok(true)
    } else {
        Ok(false)
    }
}

struct LockGroups<K: Clone> {
    groups: Vec<LockGroup<K>>,
    last: Option<LockGroup<K>>,
}
impl<K: Clone + IndexOrd> LockGroups<K> {
    fn new() -> Self {
        Self {
            groups: Default::default(),
            last: Default::default(),
        }
    }
    fn group_for_id(&mut self, pos_ref: &PosRef<K>, bottom_depth: BottomDepth) -> &mut LockGroup<K> {
        let lock_group = if let Some(lg) = self.last.take() {
            if lg.is_for(&pos_ref.node_ref) {
                lg
            } else {
                if !lg.group.is_empty() {
                    self.groups.push(lg);
                }
                LockGroup::new(bottom_depth)
            }
        } else {
            LockGroup::new(bottom_depth)
        };
        self.last = Some(lock_group);
        self.last.as_mut().unwrap()
    }

    pub fn last_set_change(&mut self, nr: &TreeNodeRef, cr: ChangeResults) {
        //TODO: replace that with a value return from the lock_logic function
        self.last.as_mut().unwrap().set_change(nr, cr)
    }
    fn finish(&mut self) {
        if let Some(lg) = self.last.take() {
            if !lg.group.is_empty() {
                self.groups.push(lg);
            }
        }
    }
}

fn find_and_change_leaf<K, V, I>(change: &KeyChanges<K, V>, updates: &mut LockGroups<K>, index: &mut I) -> PIRes<()>
where
    K: IndexTypeInternal,
    V: IndexTypeInternal,
    I: IndexModify<K, V>,
{
    let mut next: Option<PosRef<K>> = None;
    let mut path = Path::new(change.k.clone());
    let mut retry_count = 0;
    loop {
        next = if let Some(cur_node) = &mut next {
            if let Some((node_modify, version)) = index.load_modify(&cur_node.node_ref)? {
                match &*node_modify {
                    TreeNode::Node(n) => {
                        path.push(cur_node.clone(), version, node_modify.len());
                        n.find_write(&change.k)
                    }
                    TreeNode::Leaf(ref _none) => {
                        let node_ref = &cur_node.node_ref;
                        path.push(cur_node.clone(), version, node_modify.len());
                        path.compute_depths();

                        let locked = lock_logic(index, path, node_modify.clone(), updates, None)?;
                        if locked {
                            let (loaded, already_changed) = index.owned(&cur_node.node_ref, node_modify);
                            let mut leaf = loaded.as_leaf();
                            let cr = change.apply(&mut leaf, index.value_mode(), index.index_name())?;
                            if cr.add.is_some() || cr.remove.is_some() {
                                updates.last_set_change(&cur_node.node_ref, cr);
                                index.update(node_ref, TreeNode::Leaf(leaf), version)?;
                            } else if already_changed {
                                index.update(node_ref, TreeNode::Leaf(leaf), version)?;
                            }
                            break;
                        } else {
                            path = Path::new(change.k.clone());
                            None
                        }
                    }
                }
            } else {
                None
            }
        } else if let Some(r) = index.get_root_refresh()? {
            retry_count += 1;
            if retry_count > 1000000 {
                return Err(crate::IndexChangeError::ReachedLimitOfRetry);
            }
            path.clear();
            Some(PosRef::new(&change.k, 0, r))
        } else {
            retry_count += 1;
            if retry_count > 1000000 {
                return Err(crate::IndexChangeError::ReachedLimitOfRetry);
            }
            path.clear();
            index.lock_config()?;
            if let Some(r) = index.get_root_refresh()? {
                Some(PosRef::new(&change.k, 0, r))
            } else {
                let mut leaf = Leaf::new();
                change.apply(&mut leaf, index.value_mode(), index.index_name())?;
                let leaf_ref = index.insert(TreeNode::Leaf(leaf))?;
                index.set_root(Some(leaf_ref))?;
                break;
            }
        }
    }
    Ok(())
}

impl<K, V, T> IndexApply<K, V> for T
where
    K: IndexTypeInternal,
    V: IndexTypeInternal,
    T: IndexModify<K, V>,
    T: Index<K, V>,
{
    fn apply(&mut self, adds: &[KeyChanges<K, V>]) -> PIRes<()> {
        let mut changes_group = LockGroups::new();

        for add in adds {
            find_and_change_leaf(add, &mut changes_group, self)?;
        }
        changes_group.finish();

        let mut updates = changes_group.groups;

        let mut merge_ands_save = Vec::new();
        while updates.len() > 1 || (updates.len() == 1 && updates[0].group[0].path.len() > 1) {
            let (parent_updates, parent_changes) = lock_parents(self, updates)?;

            merge_ands_save.push(parent_changes);
            updates = parent_updates;
        }

        for parent_changes in merge_ands_save {
            for update in parent_changes {
                // It's locked, should not have miss read unwrap.
                rebalance(self, update)?;
            }
        }
        if updates.len() == 1 && updates[0].group[0].path.len() == 1 {
            self.lock_config()?;
            while let Some(r) = self.get_root_refresh()? {
                if let Some((n, n_version)) = self.load_modify(&r)? {
                    if n.len() > self.limits().top() {
                        let ids = split_and_save(self, r, n, n_version)?;
                        let node = TreeNode::Node(Node::new_from_split(r, &ids));
                        let cur_id = self.insert(node)?;
                        self.set_root(Some(cur_id))?;
                    } else if n.len() == 1 {
                        if let TreeNode::Node(cn) = &*n {
                            self.delete(&r, n_version)?;
                            self.set_root(Some(cn.get(0)))?;
                        } else {
                            break;
                        }
                    } else if n.len() == 0 {
                        self.set_root(None)?;
                    } else {
                        break;
                    }
                }
            }
        }
        Ok(())
    }
}

impl<K, V, T> Index<K, V> for T
where
    K: IndexTypeInternal,
    V: IndexTypeInternal,
    T: IndexKeeper<K, V>,
{
    fn get(&mut self, k: &K) -> PERes<Option<Value<V>>> {
        Ok(if let Some(node) = self.get_root()? {
            let mut cur_node = node;
            let mut reuse = None;
            loop {
                match self.load_with(&cur_node, reuse)? {
                    TreeNode::Node(n) => {
                        cur_node = n.find(k).node_ref;
                        reuse = Some(n);
                    }
                    TreeNode::Leaf(leaf) => {
                        break leaf.find(k).map(|el| el.1).ok();
                    }
                }
            }
        } else {
            None
        })
    }

    fn iter_from(&mut self, first: Bound<&K>) -> PERes<PageIter<K, V>> {
        let mut path = Vec::new();
        let mut iter = if let Some(mut cur_node) = self.get_root()? {
            path.push((0, cur_node));
            let mut reuse = None;
            loop {
                match self.load_with(&cur_node, reuse)? {
                    TreeNode::Node(n) => {
                        let value = match first {
                            Bound::Included(f) => {
                                let found = n.find(f);
                                (found.pos, found.node_ref)
                            }
                            Bound::Excluded(f) => {
                                let found = n.find(f);
                                (found.pos, found.node_ref)
                            }
                            Bound::Unbounded => (0, n.get(0)),
                        };
                        cur_node = value.1;
                        path.push(value);
                        reuse = Some(n);
                    }
                    TreeNode::Leaf(leaf) => {
                        break PageIter {
                            iter: leaf.iter_from(first).peekable(),
                        };
                    }
                }
            }
        } else {
            PageIter {
                iter: Vec::new().into_iter().peekable(),
            }
        };

        while iter.iter.peek().is_none() {
            let prev = path.pop();
            if let Some((_, node)) = path.last() {
                let (pos, _) = prev.unwrap();
                match self.load(node)? {
                    TreeNode::Node(n) => {
                        // check if there are more elements in the node
                        if n.len() > pos + 1 {
                            let mut cur_node = n.get(pos + 1);
                            loop {
                                match self.load(&cur_node)? {
                                    TreeNode::Node(n) => {
                                        cur_node = n.get(0);
                                    }
                                    TreeNode::Leaf(leaf) => {
                                        // Use here the key anyway, should start from the first in
                                        // any case
                                        iter = PageIter {
                                            iter: leaf.iter_from(first).peekable(),
                                        };
                                        break;
                                    }
                                }
                            }
                        }
                    }
                    TreeNode::Leaf(_leaf) => {
                        panic!("can't happen");
                    }
                }
            } else {
                break;
            }
        }
        Ok(iter)
    }

    fn back_iter_from(&mut self, last: Bound<&K>) -> PERes<PageIterBack<K, V>> {
        let mut path = Vec::new();
        let mut iter = if let Some(mut cur_node) = self.get_root()? {
            path.push((0, cur_node));
            let mut reuse = None;
            loop {
                match self.load_with(&cur_node, reuse)? {
                    TreeNode::Node(n) => {
                        let value = match last {
                            Bound::Included(f) => {
                                let found = n.find(f);
                                (found.pos, found.node_ref)
                            }
                            Bound::Excluded(f) => {
                                let found = n.find(f);
                                (found.pos, found.node_ref)
                            }
                            Bound::Unbounded => {
                                let pos = n.len() - 1;
                                (pos, n.get(pos))
                            }
                        };
                        cur_node = value.1;
                        path.push(value);
                        reuse = Some(n);
                    }
                    TreeNode::Leaf(leaf) => {
                        break PageIterBack {
                            iter: leaf.back_iter_from(last).peekable(),
                        };
                    }
                }
            }
        } else {
            PageIterBack {
                iter: Vec::new().into_iter().rev().peekable(),
            }
        };

        while iter.iter.peek().is_none() {
            let prev = path.pop();
            if let Some((_, node)) = path.last() {
                let (pos, _) = prev.unwrap();
                match self.load(node)? {
                    TreeNode::Node(n) => {
                        // check if there are more elements in the node
                        if pos > 0 {
                            let mut cur_node = n.get(pos - 1);
                            loop {
                                match self.load(&cur_node)? {
                                    TreeNode::Node(n) => {
                                        cur_node = n.get(n.len() - 1);
                                    }
                                    TreeNode::Leaf(leaf) => {
                                        // Use here the key anyway, should start from the first in
                                        // any case
                                        iter = PageIterBack {
                                            iter: leaf.back_iter_from(last).peekable(),
                                        };
                                        break;
                                    }
                                }
                            }
                        }
                    }
                    TreeNode::Leaf(_leaf) => {
                        panic!("can't happen");
                    }
                }
            } else {
                break;
            }
        }
        Ok(iter)
    }
}
